"""Utilities for typing"""
import typing

if hasattr(typing, "get_args"):
    _get_args = typing.get_args
else:

    def _get_args(base):
        return getattr(base, "__args__", ())


def is_generic_alias(obj) -> bool:
    for attr_name in ["GenericAlias", "_GenericAlias"]:
        if hasattr(typing, attr_name):
            if isinstance(obj, getattr(typing, attr_name)):
                return True
    return False


def get_generic_parameter(cls):
    for base in getattr(cls, "__orig_bases__", []):
        if is_generic_alias(base):
            args = _get_args(base)
            return args[0]


def get_item_cls(cls, default=None):
    param = get_generic_parameter(cls)
    if param is None or isinstance(param, typing.TypeVar):  # class is not parametrized
        return default
    return param



import pytest

from web_poet.mixins import ResponseShortcutsMixin
from web_poet.page_inputs import HttpResponse


class MyPage(ResponseShortcutsMixin):
    def __init__(self, response: HttpResponse):
        self.response = response


@pytest.fixture
def my_page(book_list_html_response):

    return MyPage(book_list_html_response)


def test_url(my_page) -> None:
    assert my_page.url == "http://books.toscrape.com/index.html"


def test_html(my_page, book_list_html) -> None:
    assert my_page.html == book_list_html


def test_xpath(my_page) -> None:
    title = my_page.xpath(".//title/text()").get().strip()
    assert title == "All products | Books to Scrape - Sandbox"


def test_css(my_page) -> None:
    title = my_page.css("title::text").get().strip()
    assert title == "All products | Books to Scrape - Sandbox"


def test_baseurl(my_page) -> None:
    assert my_page.base_url == "http://books.toscrape.com/index.html"


def test_urljoin(my_page) -> None:
    assert my_page.urljoin("foo") == "http://books.toscrape.com/foo"


def test_custom_baseurl() -> None:
    body = b"""
    <html>
    <head>
        <base href="http://example.com/foo/">
    </head>
    <body><body>
    </html>
    """
    response = HttpResponse(
        url="http://www.example.com/path",
        body=body,
    )
    page = MyPage(response=response)

    assert page.url == "http://www.example.com/path"
    assert page.base_url == "http://example.com/foo/"
    assert page.urljoin("bar") == "http://example.com/foo/bar"
    assert page.urljoin("http://example.com/1") == "http://example.com/1"



import pytest

from web_poet.exceptions import HttpError, HttpRequestError, HttpResponseError
from web_poet.page_inputs import HttpRequest, HttpResponse

URL = "https://example.com"


def test_http_error_init() -> None:
    exc = HttpError()
    assert exc.request is None
    assert exc.args

    request = HttpRequest(URL)
    exc = HttpError(request=request)
    assert exc.request == request


def test_http_request_error_init() -> None:
    exc = HttpRequestError()
    assert exc.request is None
    assert exc.args

    request = HttpRequest(URL)
    exc = HttpRequestError(request=request)
    assert exc.request == request

    response = HttpResponse(URL, b"")
    with pytest.raises(TypeError):
        HttpRequestError(request=request, response=response)  # type: ignore[call-arg]


def test_http_response_error_init() -> None:
    exc = HttpResponseError()
    assert exc.request is None
    assert exc.response is None
    assert exc.args

    request = HttpRequest(URL)
    response = HttpResponse(URL, b"")

    exc = HttpResponseError(request=request, response=response)
    assert exc.request == request
    assert exc.response == response



import os

import pytest

from web_poet.page_inputs import HttpResponse, HttpResponseBody

pytest_plugins = ["pytester"]


def read_fixture(path):
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as f:
        return f.read()


@pytest.fixture
def book_list_html():
    return read_fixture("fixtures/book_list.html")


@pytest.fixture
def book_list_html_response(book_list_html):
    body = HttpResponseBody(bytes(book_list_html, "utf-8"))
    return HttpResponse(
        url="http://books.toscrape.com/index.html", body=body, encoding="utf-8"
    )



"""
This package is just for overrides testing purposes.
"""
from typing import Any, Dict, List, Type, Union

from url_matcher import Patterns

from web_poet import ItemPage, handle_urls

# NOTE: this module contains a PO with @handle_rules
from .. import po_lib_sub  # noqa: F401


class POBase(ItemPage):
    expected_instead_of: Union[Type[ItemPage], List[Type[ItemPage]]]
    expected_patterns: Patterns
    expected_to_return: Any = None
    expected_meta: Dict[str, Any]


class POTopLevelOverriden1(ItemPage):
    ...


class POTopLevelOverriden2(ItemPage):
    ...


@handle_urls("example.com", instead_of=POTopLevelOverriden1)
@handle_urls(
    "example.com", instead_of=POTopLevelOverriden1, exclude="/*.jpg|", priority=300
)
class POTopLevel1(POBase):
    expected_instead_of = [POTopLevelOverriden1, POTopLevelOverriden1]
    expected_patterns = [
        Patterns(["example.com"], ["/*.jpg|"], priority=300),
        Patterns(["example.com"]),
    ]
    expected_to_return = [None, None]
    expected_meta = [{}, {}]  # type: ignore


@handle_urls("example.com", instead_of=POTopLevelOverriden2)
class POTopLevel2(POBase):
    expected_instead_of = POTopLevelOverriden2
    expected_patterns = Patterns(["example.com"])
    expected_to_return = None
    expected_meta = {}  # type: ignore



import operator
from pathlib import Path
from typing import Iterable, List, Optional, Set, Union

import pytest

from web_poet.testing.exceptions import (
    FieldMissing,
    FieldsUnexpected,
    FieldValueIncorrect,
    ItemValueIncorrect,
)
from web_poet.testing.fixture import OUTPUT_FILE_NAME, Fixture
from web_poet.testing.utils import comparison_error_message

# https://github.com/pytest-dev/pytest/discussions/10261
_version_tuple = getattr(pytest, "version_tuple", None)
_new_pytest = _version_tuple and _version_tuple[0] >= 7


class _PathCompatMixin:
    @property
    def _path(self):
        return self.path if _new_pytest else Path(self.fspath)


class WebPoetFile(pytest.File, _PathCompatMixin):
    """Represents a directory containing test subdirectories for one Page Object."""

    @staticmethod
    def sorted(items: List["WebPoetCollector"]) -> List["WebPoetCollector"]:
        """Sort the test list by the test name."""
        return sorted(items, key=operator.attrgetter("name"))

    def collect(self) -> Iterable[Union[pytest.Item, pytest.Collector]]:  # noqa: D102
        result: List[WebPoetCollector] = []
        path = self._path
        for entry in path.iterdir():
            if entry.is_dir():
                item: WebPoetCollector = _get_collector(
                    self,
                    name=entry.name,
                    path=entry,
                )
                if item.fixture.is_valid():
                    result.append(item)
        return self.sorted(result)


class WebPoetCollector(pytest.Collector, _PathCompatMixin):
    """Represents a directory containing one test."""

    def __init__(self, name: str, parent=None, **kwargs) -> None:
        super(WebPoetCollector, self).__init__(name, parent, **kwargs)
        self.fixture = Fixture(self._path)

    def collect(self) -> Iterable[Union[pytest.Item, pytest.Collector]]:
        """Return a list of children (items and collectors) for this
        collection node."""
        if self.config.getoption("WEB_POET_TEST_PER_ITEM", default=False):
            return [
                WebPoetItem.from_parent(parent=self, name="item", fixture=self.fixture)
            ]
        else:
            overall_tests = [
                WebPoetNoToItemException.from_parent(
                    parent=self, name="TO_ITEM_DOESNT_RAISE", fixture=self.fixture
                ),
                WebPoetNoExtraFieldsItem.from_parent(
                    parent=self, name="NO_EXTRA_FIELDS", fixture=self.fixture
                ),
            ]
            field_tests = [
                WebPoetFieldItem.from_parent(
                    parent=self, name=field, fixture=self.fixture, field_name=field
                )
                for field in self.fixture.get_expected_output_fields()
            ]
            return overall_tests + field_tests


class _WebPoetItem(pytest.Item, _PathCompatMixin):
    def __init__(self, *, fixture: Fixture, **kwargs) -> None:
        super().__init__(**kwargs)
        self.fixture = fixture


class WebPoetItem(_WebPoetItem):
    def runtest(self) -> None:
        self.fixture.assert_full_item_correct()

    def reportinfo(self):
        return self._path, 0, f"{self.fixture.short_name}"

    def repr_failure(self, excinfo, style=None):
        if isinstance(excinfo.value, ItemValueIncorrect):
            got, expected = excinfo.value.args
            return comparison_error_message(
                config=self.config,
                op="==",
                expected=expected,
                got=got,
                prefix="The output doesn't match.",
            )
        else:
            return super().repr_failure(excinfo, style)


class WebPoetNoExtraFieldsItem(_WebPoetItem):
    def runtest(self) -> None:
        if self.fixture.to_item_raised():
            raise pytest.skip(
                "Skipping a test for unexpected item fields "
                "because to_item raised an exception."
            )
        self.fixture.assert_no_extra_fields()

    def reportinfo(self):
        return self._path, 0, f"{self.fixture.short_name}: extra fields"

    def repr_failure(self, excinfo, style=None):
        if isinstance(excinfo.value, FieldsUnexpected):
            fields = excinfo.value.args[0]
            return f"The item contains unexpected fields: \n{self._format_extra_fields(fields)}"
        else:
            return super().repr_failure(excinfo, style)

    def _format_extra_fields(self, extra_fields):
        lines = []
        for field, value in extra_fields.items():
            lines.append(f" * {field} = {value!r}")
        return "\n".join(lines)


class WebPoetNoToItemException(_WebPoetItem):
    def runtest(self) -> None:
        self.fixture.assert_no_toitem_exceptions()

    def reportinfo(self):
        return (
            self._path,
            0,
            f"{self.fixture.short_name}: to_item doesn't raise an error",
        )


class WebPoetFieldItem(_WebPoetItem):
    def __init__(self, *, field_name: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.field_name = field_name

    def runtest(self) -> None:
        if self.fixture.to_item_raised():
            raise pytest.skip(
                f"Skipping a test for item.{self.field_name} "
                f"because to_item raised an exception"
            )
        self.fixture.assert_field_correct(self.field_name)

    def reportinfo(self):
        return self._path, 0, f"{self.fixture.short_name} @ {self.field_name}"

    def repr_failure(self, excinfo, style=None):
        if isinstance(excinfo.value, FieldValueIncorrect):
            got, expected = excinfo.value.args
            return comparison_error_message(
                config=self.config,
                op="==",
                expected=expected,
                got=got,
                prefix=f"item.{self.field_name} is not correct.",
            )
        elif isinstance(excinfo.value, FieldMissing):
            field_name = excinfo.value.args[0]
            return f"item.{field_name} is missing."
        else:
            return super().repr_failure(excinfo, style)


_found_type_dirs: Set[Path] = set()


def collect_file_hook(
    file_path: Path, parent: pytest.Collector
) -> Optional[pytest.Collector]:
    if file_path.name == OUTPUT_FILE_NAME:
        testcase_dir = file_path.parent
        type_dir = testcase_dir.parent
        fixture = Fixture(testcase_dir)
        if not fixture.is_valid():
            return None
        if type_dir in _found_type_dirs:
            return None
        _found_type_dirs.add(type_dir)
        file: WebPoetFile = _get_file(
            parent,
            path=type_dir,
        )
        return file
    return None


def pytest_addoption(
    parser: "pytest.Parser", pluginmanager: "pytest.PytestPluginManager"
):
    parser.addoption(
        "--web-poet-test-per-item",
        dest="WEB_POET_TEST_PER_ITEM",
        action="store_true",
        help="web-poet: use a single test per item, not a test per field",
    )


if _new_pytest:

    def _get_item(parent: pytest.Collector, *, name: str, path: Path) -> WebPoetItem:
        return WebPoetItem.from_parent(
            parent,
            name=name,
            path=path,
        )

    def _get_collector(
        parent: pytest.Collector, *, name: str, path: Path
    ) -> WebPoetCollector:
        return WebPoetCollector.from_parent(
            parent,
            name=name,
            path=path,
        )

    def _get_file(parent: pytest.Collector, *, path: Path) -> WebPoetFile:
        return WebPoetFile.from_parent(
            parent,
            path=path,
        )

    def pytest_collect_file(
        file_path: Path, parent: pytest.Collector
    ) -> Optional[pytest.Collector]:
        return collect_file_hook(file_path, parent)

else:
    import py.path

    def _get_item(parent: pytest.Collector, *, name: str, path: Path) -> WebPoetItem:
        return WebPoetItem.from_parent(
            parent,
            name=name,
        )

    def _get_collector(
        parent: pytest.Collector, *, name: str, path: Path
    ) -> WebPoetCollector:
        return WebPoetCollector.from_parent(
            parent,
            name=name,
            fspath=py.path.local(path),
        )

    def _get_file(parent: pytest.Collector, *, path: Path) -> WebPoetFile:
        return WebPoetFile.from_parent(
            parent,
            fspath=py.path.local(path),
        )

    def pytest_collect_file(  # type: ignore[misc]
        path: py.path.local, parent: pytest.Collector
    ) -> Optional[pytest.Collector]:
        return collect_file_hook(Path(path), parent)



from typing import Callable, Union
from unittest import mock

import pytest

from web_poet.exceptions import HttpResponseError, RequestDownloaderVarError
from web_poet.page_inputs import (
    HttpClient,
    HttpRequest,
    HttpRequestBody,
    HttpRequestHeaders,
    HttpResponse,
)
from web_poet.requests import request_downloader_var


@pytest.fixture
def async_mock():
    """Workaround since python 3.7 doesn't ship with asyncmock."""

    async def async_test(req):
        return HttpResponse(str(req.url), body=b"")

    mock.MagicMock.__await__ = lambda x: async_test(x).__await__()

    return async_test


@pytest.mark.asyncio
async def test_perform_request_from_httpclient(async_mock) -> None:

    url = "http://example.com"
    client = HttpClient()

    with pytest.raises(RequestDownloaderVarError):
        await client.get(url)

    request_downloader_var.set(async_mock)
    response = await client.get(url)

    # The async downloader implementation should return the HttpResponse
    assert str(response.url) == str(url)
    assert isinstance(response, HttpResponse)


@pytest.mark.asyncio
async def test_http_client_single_requests(async_mock) -> None:
    client = HttpClient(async_mock)

    with mock.patch("web_poet.page_inputs.client.HttpRequest") as mock_request:
        await client.request("url")
        await client.get("url-get", headers={"X-Headers": "123"})
        await client.post("url-post", headers={"X-Headers": "123"}, body=b"body value")
        assert mock_request.call_args_list == [
            mock.call(
                url="url",
                method="GET",
                headers=HttpRequestHeaders(),
                body=HttpRequestBody(),
            ),
            mock.call(
                url="url-get",
                method="GET",
                headers=HttpRequestHeaders({"X-Headers": "123"}),
                body=HttpRequestBody(),
            ),
            mock.call(
                url="url-post",
                method="POST",
                headers=HttpRequestHeaders({"X-Headers": "123"}),
                body=HttpRequestBody(b"body value"),
            ),
        ]


@pytest.fixture
def client_with_status() -> Callable:
    def _param_wrapper(status_code: int):
        async def stub_request_downloader(*args, **kwargs):
            async def stub(req):
                return HttpResponse(req.url, body=b"", status=status_code)

            return await stub(*args, **kwargs)

        return stub_request_downloader

    return _param_wrapper


@pytest.mark.asyncio
@pytest.mark.parametrize("method_name", ["request", "get", "post", "execute"])
async def test_http_client_allow_status(
    async_mock, client_with_status, method_name
) -> None:
    client = HttpClient(async_mock)

    # Simulate 500 Internal Server Error responses
    client._request_downloader = client_with_status(500)

    method = getattr(client, method_name)

    url_or_request: Union[str, HttpRequest] = "url"
    if method_name == "execute":
        # NOTE: We're ignoring the type below due to the following mypy bugs:
        #   - https://github.com/python/mypy/issues/10187
        #   - https://github.com/python/mypy/issues/5313
        #   - https://github.com/python-attrs/attrs/issues/889
        # Currently, the said bugs causes mypy to raise the following error:
        #   'Incompatible types in assignment (expression has type "ResponseUrl",
        #   variable has type "Optional[str]")'
        url_or_request = HttpRequest(url_or_request)  # type: ignore[arg-type]

    # Should handle single and multiple values
    await method(url_or_request, allow_status=500)
    response = await method(url_or_request, allow_status=[500, 503])
    assert isinstance(response, HttpResponse)
    assert response.status == 500

    # As well as strings
    await method(url_or_request, allow_status="500")
    await method(url_or_request, allow_status=["500", "503"])

    with pytest.raises(HttpResponseError) as excinfo:
        await method(url_or_request)
    assert isinstance(excinfo.value.request, HttpRequest)
    assert isinstance(excinfo.value.response, HttpResponse)
    assert str(excinfo.value).startswith("500 INTERNAL_SERVER_ERROR response for")

    with pytest.raises(HttpResponseError):
        await method(url_or_request, allow_status=406)
    assert isinstance(excinfo.value.request, HttpRequest)
    assert isinstance(excinfo.value.response, HttpResponse)
    assert str(excinfo.value).startswith("500 INTERNAL_SERVER_ERROR response for")

    # As long as "*" is present, then no errors would be raised
    await method(url_or_request, allow_status="*")
    await method(url_or_request, allow_status=[500, "*"])

    # Globbing isn't supported
    with pytest.raises(HttpResponseError):
        await method(url_or_request, allow_status="5*")
    assert isinstance(excinfo.value.request, HttpRequest)
    assert isinstance(excinfo.value.response, HttpResponse)
    assert str(excinfo.value).startswith("500 INTERNAL_SERVER_ERROR response for")


@pytest.mark.asyncio
async def test_http_client_keyword_enforcing(async_mock) -> None:
    """Only keyword args are allowed after the url param."""

    client = HttpClient(async_mock)

    with pytest.raises(TypeError):
        await client.request("url", "PATCH")  # type: ignore[misc]

    with pytest.raises(TypeError):
        await client.get("url", {"Content-Encoding": "utf-8"})  # type: ignore[misc]

    with pytest.raises(TypeError):
        await client.post("url", {"X-Header": "value"}, b"body")  # type: ignore[misc]


@pytest.mark.asyncio
async def test_http_client_execute(async_mock) -> None:
    client = HttpClient(async_mock)

    request = HttpRequest("url-1")
    response = await client.execute(request)

    assert isinstance(response, HttpResponse)
    assert str(response.url) == "url-1"


@pytest.mark.asyncio
async def test_http_client_batch_execute(async_mock) -> None:
    client = HttpClient(async_mock)

    requests = [
        HttpRequest("url-1"),
        HttpRequest("url-get", method="GET"),
        HttpRequest("url-post", method="POST"),
    ]
    responses = await client.batch_execute(*requests)

    assert all([isinstance(response, HttpResponse) for response in responses])


@pytest.fixture
def client_that_errs(async_mock) -> HttpClient:
    client = HttpClient(async_mock)

    # Simulate errors inside the request coroutines
    async def stub_request_downloader(*args, **kwargs):
        async def err():
            raise ValueError("test exception")

        return await err()

    client._request_downloader = stub_request_downloader

    return client


@pytest.mark.asyncio
async def test_http_client_batch_execute_with_exception(client_that_errs) -> None:

    requests = [
        HttpRequest("url-1"),
        HttpRequest("url-get", method="GET"),
        HttpRequest("url-post", method="POST"),
    ]
    responses = await client_that_errs.batch_execute(*requests, return_exceptions=True)

    assert len(responses) == 3
    assert isinstance(responses[0], Exception)
    assert isinstance(responses[1], Exception)
    assert isinstance(responses[2], Exception)


@pytest.mark.asyncio
async def test_http_client_batch_execute_with_exception_raised(
    client_that_errs,
) -> None:
    requests = [
        HttpRequest("url-1"),
    ]
    with pytest.raises(ValueError):
        await client_that_errs.batch_execute(*requests)


@pytest.mark.asyncio
async def test_http_client_batch_execute_allow_status(
    async_mock, client_with_status
) -> None:
    client = HttpClient(async_mock)

    # Simulate 400 Bad Request
    client._request_downloader = client_with_status(400)

    requests = [HttpRequest("url-1"), HttpRequest("url-2"), HttpRequest("url-3")]

    await client.batch_execute(*requests, allow_status=400)
    await client.batch_execute(*requests, allow_status=[400, 403])
    await client.batch_execute(*requests, allow_status="400")
    responses = await client.batch_execute(*requests, allow_status=["400", "403"])

    for r in responses:
        assert isinstance(r, HttpResponse) and r.status == 400

    with pytest.raises(HttpResponseError) as excinfo:
        await client.batch_execute(*requests)
    assert isinstance(excinfo.value.request, HttpRequest)
    assert isinstance(excinfo.value.response, HttpResponse)
    assert str(excinfo.value).startswith("400 BAD_REQUEST response for")

    with pytest.raises(HttpResponseError) as excinfo:
        await client.batch_execute(*requests, allow_status=406)
    assert isinstance(excinfo.value.request, HttpRequest)
    assert isinstance(excinfo.value.response, HttpResponse)
    assert str(excinfo.value).startswith("400 BAD_REQUEST response for")

    await client.batch_execute(*requests, return_exceptions=True, allow_status=400)
    await client.batch_execute(
        *requests, return_exceptions=True, allow_status=[400, 403]
    )
    await client.batch_execute(*requests, return_exceptions=True, allow_status="400")
    await client.batch_execute(
        *requests, return_exceptions=True, allow_status=["400", "403"]
    )

    responses = await client.batch_execute(*requests, return_exceptions=True)
    for r in responses:
        assert (
            isinstance(r, HttpResponseError)
            and isinstance(r.request, HttpRequest)
            and isinstance(r.response, HttpResponse)
        )
    assert all([str(r).startswith("400 BAD_REQUEST response for") for r in responses])

    responses = await client.batch_execute(
        *requests, return_exceptions=True, allow_status=408
    )
    for r in responses:
        assert (
            isinstance(r, HttpResponseError)
            and isinstance(r.request, HttpRequest)
            and isinstance(r.response, HttpResponse)
        )
    assert all([str(r).startswith("400 BAD_REQUEST response for") for r in responses])

    # These have no assertions since they're used to see if mypy raises an
    # error against them.
    for r in responses:
        if isinstance(r, HttpResponseError):
            r.request
            r.response
        else:
            r.url
            r.body
            r.status
            r.headers



"""This package is being used by tests/po_lib to validate some behaviors on
external depedencies.
"""
from typing import Any, Dict, Type

from url_matcher import Patterns

from web_poet import ItemPage, handle_urls


class POBase(ItemPage):
    expected_instead_of: Type[ItemPage]
    expected_patterns: Patterns
    expected_meta: Dict[str, Any]


class POLibSubOverriden(ItemPage):
    ...


@handle_urls("sub.example", instead_of=POLibSubOverriden)
class POLibSub(POBase):
    expected_instead_of = POLibSubOverriden
    expected_patterns = Patterns(["sub.example"])
    expected_to_return = None
    expected_meta = {}  # type: ignore



import abc
import typing

import attr

from web_poet._typing import get_item_cls
from web_poet.fields import FieldsMixin, item_from_fields
from web_poet.mixins import ResponseShortcutsMixin
from web_poet.page_inputs import HttpResponse
from web_poet.utils import _create_deprecated_class


class Injectable(abc.ABC, FieldsMixin):
    """Base Page Object class, which all Page Objects should inherit from
    (probably through Injectable subclasses).

    Frameworks which are using ``web-poet`` Page Objects should use
    :func:`is_injectable` function to detect if an object is an Injectable,
    and if an object is injectable, allow building it automatically
    through dependency injection, using https://github.com/scrapinghub/andi
    library.

    Instead of inheriting you can also use ``Injectable.register(MyWebPage)``.
    ``Injectable.register`` can also be used as a decorator.
    """

    pass


# NoneType is considered as injectable. Required for Optionals to work.
Injectable.register(type(None))


def is_injectable(cls: typing.Any) -> bool:
    """Return True if ``cls`` is a class which inherits
    from :class:`~.Injectable`."""
    return isinstance(cls, type) and issubclass(cls, Injectable)


ItemT = typing.TypeVar("ItemT")


class Returns(typing.Generic[ItemT]):
    """Inherit from this generic mixin to change the item class used by
    :class:`~.ItemPage`"""

    @property
    def item_cls(self) -> typing.Type[ItemT]:
        """Item class"""
        return get_item_cls(self.__class__, default=dict)


class ItemPage(Injectable, Returns[ItemT]):
    """Base Page Object, with a default :meth:`to_item` implementation
    which supports web-poet fields.
    """

    _skip_nonitem_fields: bool

    def __init_subclass__(cls, skip_nonitem_fields: bool = False, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._skip_nonitem_fields = skip_nonitem_fields

    async def to_item(self) -> ItemT:
        """Extract an item from a web page"""
        return await item_from_fields(
            self, item_cls=self.item_cls, skip_nonitem_fields=self._skip_nonitem_fields
        )


@attr.s(auto_attribs=True)
class WebPage(ItemPage[ItemT], ResponseShortcutsMixin):
    """Base Page Object which requires :class:`~.HttpResponse`
    and provides XPath / CSS shortcuts.
    """

    response: HttpResponse


ItemWebPage = _create_deprecated_class("ItemWebPage", WebPage, warn_once=False)



from attrs import define

from web_poet import ItemPage, default_registry, field, handle_urls
from web_poet.example import get_item


def _revert_add_rule():
    default_registry._rule_counter -= 1
    rule_id = list(default_registry._rules)[-1]
    default_registry._rules.pop(rule_id)
    default_registry._overrides_matchers[None].remove(rule_id)
    default_registry._item_matchers[None].remove(rule_id)


def test_async_to_item():
    @define
    class Item:
        foo: str

    try:

        @handle_urls("")
        class Page(ItemPage[Item]):
            @field
            async def foo(self):
                return "bar"

        item = get_item("file:///dev/null", Item)
        assert isinstance(item, Item)
        assert item.foo == "bar"
    finally:
        _revert_add_rule()


def test_sync_to_item():
    @define
    class Item:
        foo: str

    try:

        @handle_urls("")
        class Page(ItemPage[Item]):
            def to_item(self):
                return Item(foo="bar")

        item = get_item("file:///dev/null", Item)
        assert isinstance(item, Item)
        assert item.foo == "bar"
    finally:
        _revert_add_rule()



import pytest

from web_poet import RequestUrl, ResponseUrl
from web_poet.page_inputs.url import _Url


def test_url_base_class() -> None:
    url_str = "http://example.com"
    url = _Url(url_str)
    assert str(url) == url_str
    assert repr(url) == "_Url('http://example.com')"


def test_url_init_validation() -> None:
    with pytest.raises(TypeError):
        _Url(123)  # type: ignore[arg-type]


def test_url_subclasses() -> None:
    url_str = "http://example.com"

    class MyUrl(_Url):
        pass

    class MyUrl2(_Url):
        pass

    url = MyUrl(url_str)
    assert str(url) == url_str
    assert url._url == url_str
    assert repr(url) == "MyUrl('http://example.com')"

    url2 = MyUrl2(url)
    assert str(url2) == str(url)


@pytest.mark.parametrize("url_cls", [_Url, RequestUrl, ResponseUrl])
def test_str_equality(url_cls) -> None:
    url_str = "http://example.com#foo"
    url = url_cls(url_str)
    assert url != url_str
    assert str(url) == url_str


def test_url_classes_eq() -> None:
    url_str = "http://example.com#foo"
    request_url = RequestUrl(url_str)
    response_url = ResponseUrl(url_str)

    assert request_url != response_url
    assert str(request_url) == str(response_url)



from . import functions  # needed to run register functions
from .api import (
    DeserializeFunction,
    SerializedData,
    SerializedDataFileStorage,
    SerializedLeafData,
    SerializeFunction,
    deserialize,
    deserialize_leaf,
    load_class,
    register_serialization,
    serialize,
    serialize_leaf,
)



from web_poet.mixins import SelectableMixin


class BrowserHtml(SelectableMixin, str):
    """HTML returned by a web browser,
    i.e. snapshot of the DOM tree in HTML format.
    """

    def _selector_input(self) -> str:
        return self



import asyncio
import datetime
import json
import logging
import os
import sys
from pathlib import Path
from typing import Any, Iterable, Optional, Type, TypeVar, Union

import dateutil.parser
import dateutil.tz
import time_machine
from itemadapter import ItemAdapter

from web_poet import ItemPage
from web_poet.serialization import (
    SerializedDataFileStorage,
    deserialize,
    load_class,
    serialize,
)
from web_poet.utils import ensure_awaitable, memoizemethod_noargs

from .exceptions import (
    FieldMissing,
    FieldsUnexpected,
    FieldValueIncorrect,
    ItemValueIncorrect,
)

logger = logging.getLogger(__name__)


INPUT_DIR_NAME = "inputs"
OUTPUT_FILE_NAME = "output.json"
META_FILE_NAME = "meta.json"


FixtureT = TypeVar("FixtureT", bound="Fixture")


def _get_available_filename(template: str, directory: Union[str, os.PathLike]) -> str:
    i = 1
    while True:
        result = Path(directory, template.format(i))
        if not result.exists():
            return result.name
        i += 1


class Fixture:
    """Represents a directory containing one test."""

    def __init__(self, path: Path) -> None:
        self.path = path
        self._output_error: Optional[Exception] = None

    @property
    def type_name(self) -> str:
        """The name of the type being tested."""
        return self.path.parent.name

    @property
    def test_name(self) -> str:
        """The name of the test."""
        return self.path.name

    @property
    def short_name(self) -> str:
        """The name of this fixture"""
        return f"{self.type_name}/{self.test_name}"

    @property
    def input_path(self) -> Path:
        """The inputs subdirectory path."""
        return self.path / INPUT_DIR_NAME

    @property
    def output_path(self) -> Path:
        """The output file path."""
        return self.path / OUTPUT_FILE_NAME

    @property
    def meta_path(self) -> Path:
        """The metadata file path."""
        return self.path / META_FILE_NAME

    def is_valid(self) -> bool:
        """Return True if the fixture file structure is correct, False otherwise."""
        return self.input_path.is_dir() and self.output_path.is_file()

    def get_page(self) -> ItemPage:
        """Return the page object created from the saved input."""
        cls = load_class(self.type_name)
        if not issubclass(cls, ItemPage):
            raise TypeError(f"{self.type_name} is not a descendant of ItemPage")
        storage = SerializedDataFileStorage(self.input_path)
        return deserialize(cls, storage.read())

    def get_meta(self) -> dict:
        """Return the test metadata."""
        if not self.meta_path.exists():
            return {}
        return json.loads(self.meta_path.read_bytes())

    def _get_output(self) -> dict:
        page = self.get_page()
        item = asyncio.run(ensure_awaitable(page.to_item()))
        return ItemAdapter(item).asdict()

    @memoizemethod_noargs
    def get_output(self) -> dict:
        """
        Return the output from the recreated Page Object,
        taking frozen time in account.
        """
        try:
            meta = self.get_meta()
            frozen_time: Optional[str] = meta.get("frozen_time")
            if frozen_time:
                frozen_time_parsed = self._parse_frozen_time(frozen_time)
                with time_machine.travel(frozen_time_parsed):
                    return self._get_output()
            else:
                return self._get_output()
        except Exception as e:
            self._output_error = e
            raise

    @classmethod
    def item_to_json(cls, item: Any) -> str:
        """Convert an item to a JSON string."""
        return json.dumps(ItemAdapter(item).asdict(), ensure_ascii=False, indent=4)

    @memoizemethod_noargs
    def get_expected_output(self) -> dict:
        """Return the saved output."""
        return json.loads(self.output_path.read_bytes())

    @staticmethod
    def _parse_frozen_time(meta_value: str) -> datetime.datetime:
        """Parse and possibly fix the frozen_time metadata string."""
        parsed_value = dateutil.parser.parse(meta_value)

        if parsed_value.tzinfo is None:
            # if it's left as None, time_machine will set it to timezone.utc,
            # but we want to interpret the value as local time
            parsed_value = parsed_value.astimezone()
            return parsed_value

        if not time_machine.HAVE_TZSET:
            logger.warning(
                f"frozen_time {meta_value} includes timezone data which"
                f" is not supported on Windows, converting to local"
            )
            return parsed_value.astimezone()

        if sys.version_info >= (3, 9):
            from zoneinfo import ZoneInfo
        else:
            from backports.zoneinfo import ZoneInfo

        if parsed_value.tzinfo == dateutil.tz.UTC:
            return parsed_value.replace(tzinfo=ZoneInfo("UTC"))

        offset = parsed_value.tzinfo.utcoffset(None)
        assert offset is not None  # typing
        offset_hours = int(offset.days * 24 + offset.seconds / 3600)
        tzinfo = ZoneInfo(f"Etc/GMT{-offset_hours:+d}")
        return parsed_value.replace(tzinfo=tzinfo)

    def get_expected_output_fields(self):
        """Return a list of the expected output field names."""
        output = self.get_expected_output()
        return list(output.keys())

    def assert_full_item_correct(self):
        """Get the output and assert that it matches the expected output."""
        output = self.get_output()
        expected_output = self.get_expected_output()
        if output != expected_output:
            raise ItemValueIncorrect(output, expected_output)

    def assert_field_correct(self, name: str):
        """Assert that a certain field in the output matches the expected value"""
        expected_output = self.get_expected_output()[name]
        if name not in self.get_output():
            raise FieldMissing(name)
        output = self.get_output()[name]
        if output != expected_output:
            raise FieldValueIncorrect(output, expected_output)

    def assert_no_extra_fields(self):
        """Assert that there are no extra fields in the output"""
        output = self.get_output()
        expected_output = self.get_expected_output()
        extra_field_keys = output.keys() - expected_output.keys()
        extra_fields = {key: output[key] for key in extra_field_keys}
        if extra_fields:
            raise FieldsUnexpected(extra_fields)

    def to_item_raised(self) -> bool:
        """Return True if to_item raised an error.
        Note that if to_item hasn't been called yet, this method returns False.
        """
        return self._output_error is not None

    def assert_no_toitem_exceptions(self):
        """Assert that to_item() can be run (doesn't raise an error)"""
        self.get_output()

    @classmethod
    def save(
        cls: Type[FixtureT],
        base_directory: Union[str, os.PathLike],
        *,
        inputs: Iterable[Any],
        item: Any,
        meta: Optional[dict] = None,
        fixture_name=None,
    ) -> FixtureT:
        """Save and return a fixture."""
        if not fixture_name:
            fixture_name = _get_available_filename(
                "test-{}", base_directory  # noqa: P103
            )
        fixture_dir = Path(base_directory, fixture_name)
        fixture = cls(fixture_dir)
        fixture.input_path.mkdir(parents=True)

        serialized_inputs = serialize(inputs)
        storage = SerializedDataFileStorage(fixture.input_path)
        storage.write(serialized_inputs)

        with fixture.output_path.open("w") as f:
            f.write(cls.item_to_json(item))

        if meta:
            with fixture.meta_path.open("w") as f:
                json.dump(meta, f, ensure_ascii=False, indent=4)

        return fixture



from .fields import field, item_from_fields, item_from_fields_sync
from .page_inputs import (
    BrowserHtml,
    HttpClient,
    HttpRequest,
    HttpRequestBody,
    HttpRequestHeaders,
    HttpResponse,
    HttpResponseBody,
    HttpResponseHeaders,
    PageParams,
    RequestUrl,
    ResponseUrl,
)
from .pages import Injectable, ItemPage, ItemWebPage, Returns, WebPage
from .requests import request_downloader_var
from .rules import (
    ApplyRule,
    OverrideRule,
    PageObjectRegistry,
    RulesRegistry,
    consume_modules,
)
from .utils import cached_method

default_registry = RulesRegistry()
handle_urls = default_registry.handle_urls



import json
from typing import Dict, List, Type

from .. import (
    HttpClient,
    HttpRequest,
    HttpRequestBody,
    HttpResponse,
    HttpResponseBody,
    PageParams,
)
from ..page_inputs.client import _SavedResponseData
from ..page_inputs.url import _Url
from .api import (
    SerializedLeafData,
    deserialize_leaf,
    register_serialization,
    serialize_leaf,
)


def _serialize_HttpRequest(o: HttpRequest) -> SerializedLeafData:
    info = {
        "url": str(o.url),
        "method": o.method,
        "headers": list(o.headers.items()),
    }
    result: SerializedLeafData = {
        "info.json": json.dumps(
            info, ensure_ascii=False, sort_keys=True, indent=2
        ).encode(),
    }
    if o.body:
        result["body.txt"] = bytes(o.body)
    return result


def _deserialize_HttpRequest(
    cls: Type[HttpRequest], data: SerializedLeafData
) -> HttpRequest:
    body = HttpRequestBody(data.get("body.txt", b""))
    info = json.loads(data["info.json"])
    return cls(
        body=body,
        url=info["url"],
        method=info["method"],
        headers=info["headers"],
    )


register_serialization(_serialize_HttpRequest, _deserialize_HttpRequest)


def _serialize_HttpResponse(o: HttpResponse) -> SerializedLeafData:
    info = {
        "url": str(o.url),
        "status": o.status,
        "headers": list(o.headers.items()),
        "_encoding": o._encoding,
    }
    return {
        "body.html": bytes(o.body),
        "info.json": json.dumps(
            info, ensure_ascii=False, sort_keys=True, indent=2
        ).encode(),
    }


def _deserialize_HttpResponse(
    cls: Type[HttpResponse], data: SerializedLeafData
) -> HttpResponse:
    body = HttpResponseBody(data["body.html"])
    info = json.loads(data["info.json"])
    return cls(
        body=body,
        url=info["url"],
        status=info["status"],
        headers=info["headers"],
        encoding=info["_encoding"],
    )


register_serialization(_serialize_HttpResponse, _deserialize_HttpResponse)


def _serialize_HttpResponseBody(o: HttpResponseBody) -> SerializedLeafData:
    return {"html": bytes(o)}


def _deserialize_HttpResponseBody(
    cls: Type[HttpResponseBody], data: SerializedLeafData
) -> HttpResponseBody:
    return cls(data["html"])


register_serialization(_serialize_HttpResponseBody, _deserialize_HttpResponseBody)


def _serialize__Url(o: _Url) -> SerializedLeafData:
    return {"txt": str(o).encode()}


def _deserialize__Url(cls: Type[_Url], data: SerializedLeafData) -> _Url:
    return cls(data["txt"].decode())


register_serialization(_serialize__Url, _deserialize__Url)


def _serialize_HttpClient(o: HttpClient) -> SerializedLeafData:
    serialized_data: SerializedLeafData = {
        "exists": b"",
    }
    for i, data in enumerate(o.get_saved_responses()):
        serialized_request = serialize_leaf(data.request)
        for k, v in serialized_request.items():
            serialized_data[f"{i}-HttpRequest.{k}"] = v
        serialized_response = serialize_leaf(data.response)
        for k, v in serialized_response.items():
            serialized_data[f"{i}-HttpResponse.{k}"] = v
    return serialized_data


def _deserialize_HttpClient(
    cls: Type[HttpClient], data: SerializedLeafData
) -> HttpClient:
    responses: List[_SavedResponseData] = []

    serialized_requests: Dict[str, SerializedLeafData] = {}
    serialized_responses: Dict[str, SerializedLeafData] = {}
    for k, v in data.items():
        if k == "exists":
            continue
        # k is number-("HttpRequest"|"HttpResponse").("body"|"info").ext
        key, type_suffix = k.split("-", 1)
        type_name, suffix = type_suffix.split(".", 1)
        if type_name == "HttpRequest":
            serialized_requests.setdefault(key, {})[suffix] = v
        elif type_name == "HttpResponse":
            serialized_responses.setdefault(key, {})[suffix] = v

    for key, serialized_request in serialized_requests.items():
        serialized_response = serialized_responses.get(key)
        if not serialized_response:
            continue
        request = deserialize_leaf(HttpRequest, serialized_request)
        response = deserialize_leaf(HttpResponse, serialized_response)
        responses.append(_SavedResponseData(request, response))

    return cls(return_only_saved_responses=True, responses=responses)


register_serialization(_serialize_HttpClient, _deserialize_HttpClient)


def _serialize_PageParams(o: PageParams) -> SerializedLeafData:
    return {
        "json": json.dumps(o, ensure_ascii=False, sort_keys=True, indent=2).encode()
    }


def _deserialize_PageParams(
    cls: Type[PageParams], data: SerializedLeafData
) -> PageParams:
    return cls(json.loads(data["json"]))


register_serialization(_serialize_PageParams, _deserialize_PageParams)
